# Module containing frequently used tools already implemented.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import matplotlib.pyplot as plt
from .base import ScoreScaler


class MinMaxScaler(ScoreScaler):
    """
    Class that performs Min-Max scaling and scales the values to the range defined by
    [min_val, max_val].

    Parameters
    ----------
    min_val: float, default=0.0
        Minimum value to scale values to.

    max_val: float, default=1.0
        Maximum value to scale values to.

    max_historic: bool, default=False
        Parameter indicating whether to use the best historical score as the maximum value for
        Min-Max scaling.
    """
    def __init__(self, min_val: float = 0.0, max_val: float = 1.0, max_historic: bool = False):
        self._min_val = min_val
        self._max_val = max_val
        self._max_historic = max_historic

    def __repr__(self):
        return f'MinMaxScaler(min_val={self._min_val}, max_val={self._max_val}, ' \
               f'max_historic={self._max_historic})'

    def scale(self, ant_scores: np.ndarray, best_historic: float):
        min_score = np.min(ant_scores)
        max_score = best_historic if self._max_historic else np.max(ant_scores)

        if min_score == max_score:  # To avoid zero division error
            return np.ones(shape=ant_scores.shape)

        return ((ant_scores - min_score) / (max_score - min_score)) * \
               (self._max_val - self._min_val) + self._min_val


class HyperparameterCheckerAS(object):
    """
    Class to plot the effects of the hyperparameters associated with the Ant System (AS) pheromone
    updating strategy on pheromone updating.

    The plot() method will plot two graphs, the first one representing the increase in the values
    of the pheromone matrix assuming that all ants used for its update are used, i.e. in a real
    case all ants that have visited the same connection, and the second one the decrease in the
    value of the pheromone matrix when no ant visits a given connection during all the plotted
    interactions.
    In the increment graph, several lines will be plotted, each corresponding to the random
    assignment of a score to each ant in the range of values indicated in the legend.

    Parameters
    ----------
    n_ants: int
        Number of ants used for updating the pheromone matrix.

    pher_init: float
        Initial value of the pheromone matrix.

    evaporation: float
        Pheromone evaporation parameter.

    weight: float, default=1.0
        Weight used to model the contribution of each ant to the pheromone update (it is assumed
        that the ants' scores have been scaled to the range [0,1]).

    seed: int, default=None
        Random seed.

    Methods
    -------
    plot(num_iterations: int = 100, figsize: tuple = (10, 10), linewidth: float = 2.5,
        title_size: int = 15, save_plot: str = None):

        Method to plotting the increase and decrease of pheromone matrix values over specified
        interactions.
    """
    def __init__(self, n_ants: int, pher_init: float, evaporation: float, weight: float = 1.0,
                 seed: int = None):
        self.n_ants = n_ants
        self.pher_init = pher_init
        self.evaporation = evaporation
        self.weight = weight
        self.seed = seed

    def plot(self, num_iterations: int = 100, figsize: tuple = (10, 10), linewidth: float = 2.5,
             title_size: int = 15, save_plot: str = None):
        """
        Parameters
        ----------
        num_iterations: int, default=100
            Number of interactions to simulate.

        figsize: tuple, default=(10, 10)
            Tuple indicating the size of the figure.

        linewidth: float, default=2.5
            Thickness of the lines in the plot.

        title_size: int, default=15
            Size of the title of each graphic.

        save_plot: str, default=None
            File in which to save the generated graph, if no value is provided the graph will not
            be saved.
        """
        iterations = [it for it in range(num_iterations)]
        pher_increase_100 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            weight=self.weight, iterations=iterations, score_range=(0.99, 1.0), seed=self.seed)
        pher_increase_75 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            weight=self.weight, iterations=iterations, score_range=(0.75, 1.0), seed=self.seed)
        pher_increase_50 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            weight=self.weight, iterations=iterations, score_range=(0.50, 1.0), seed=self.seed)
        pher_increase_25 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            weight=self.weight, iterations=iterations, score_range=(0.25, 1.0), seed=self.seed)
        pher_decrease = self._pherDecrease(
            pher_init=self.pher_init, evaporation=self.evaporation, iterations=iterations)

        fig, axes = plt.subplots(2, figsize=figsize)
        axes[0].plot(iterations, pher_increase_100, label='Scores [0.99-1.0]', color='#641E16',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_75, label='Scores [0.75-1.0]', color='#A93226',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_50, label='Scores [0.50-1.0]', color='#D98880',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_25, label='Scores [0.25-1.0]', color='#F2D7D5',
                     linewidth=linewidth)
        axes[1].plot(iterations, pher_decrease, color='#2980B9', linewidth=linewidth)

        axes[0].set_title('Pheromone increase', fontsize=title_size)
        axes[0].spines['top'].set_visible(False)
        axes[0].spines['right'].set_visible(False)
        axes[0].legend()
        axes[1].set_title('Pheromone decrease', fontsize=title_size)
        axes[1].spines['top'].set_visible(False)
        axes[1].spines['right'].set_visible(False)

        if save_plot is not None:
            plt.savefig(save_plot, dpi=150)

        plt.show()

    @staticmethod
    def _pherIncrease(n_ants: int, pher_init: float, evaporation: float, weight: float,
                      iterations: list, score_range: tuple, seed: int):
        """ Simulate the pheromone update (increasing the values) according to the AS update
        strategy. """
        if seed is not None:
            np.random.seed(seed)
        pher_increase = []
        pher = pher_init
        for it in iterations:
            score = np.sum(
                np.random.uniform(low=score_range[0], high=score_range[1], size=n_ants) * weight)
            pher = (1 - evaporation) * pher + score * weight
            pher_increase.append(pher)

        return pher_increase

    @staticmethod
    def _pherDecrease(pher_init: float, evaporation: float, iterations: list):
        """ Simulate the pheromone update (decreasing the values) according to the AS update
        strategy. """
        pher_decrease = []
        pher = pher_init
        for it in iterations:
            pher = (1 - evaporation) * pher
            pher_decrease.append(pher)

        return pher_decrease


class HyperparameterCheckerMMAS(object):
    """
    Class to plot the effects of the hyperparameters associated with the Min-Max Ant System (MMAS)
    pheromone updating strategy on pheromone updating.

    The plot() method will plot two graphs, the first one representing the increase in the values
    of the pheromone matrix assuming that all ants used for its update are used, i.e. in a real
    case all ants that have visited the same connection, and the second one the decrease in the
    value of the pheromone matrix when no ant visits a given connection during all the plotted
    interactions.
    In the increment graph, several lines will be plotted, each corresponding to the random
    assignment of a score to each ant in the range of values indicated in the legend.

    Parameters
    ----------
    n_ants: int
        Number of ants used for updating the pheromone matrix.

    pher_init: float
        Initial value of the pheromone matrix.

    evaporation: float
        Pheromone evaporation parameter.

    limits: tuple
        Bounds of the MMAS update strategy.

    weight: float, default=1.0
        Weight used to model the contribution of each ant to the pheromone update (it is assumed
        that the ants' scores have been scaled to the range [0,1]).

    seed: int, default=None
        Random seed.

    Methods
    -------
    plot(num_iterations: int = 100, figsize: tuple = (10, 10), linewidth: float = 2.5,
        title_size: int = 15, save_plot: str = None):

        Method to plotting the increase and decrease of pheromone matrix values over specified
        interactions.
    """
    def __init__(self, n_ants: int, pher_init: float, evaporation: float, limits: tuple,
                 weight: float = 1.0, seed: int = None):
        self.n_ants = n_ants
        self.pher_init = pher_init
        self.evaporation = evaporation
        self.limits = limits
        self.weight = weight
        self.seed = seed

    def plot(self, num_iterations: int = 100, figsize: tuple = (10, 10), linewidth: float = 2.5,
             title_size: int = 15, save_plot: str = None):
        """
        Parameters
        ----------
        num_iterations: int, default=100
            Number of interactions to simulate.

        figsize: tuple, default=(10, 10)
            Tuple indicating the size of the figure.

        linewidth: float, default=2.5
            Thickness of the lines in the plot.

        title_size: int, default=15
            Size of the title of each graphic.

        save_plot: str, default=None
            File in which to save the generated graph, if no value is provided the graph will not
            be saved.
        """
        iterations = [it for it in range(num_iterations)]
        pher_increase_100 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            limits=self.limits, weight=self.weight, iterations=iterations, score_range=(0.99, 1.0),
            seed=self.seed)
        pher_increase_75 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            limits=self.limits, weight=self.weight, iterations=iterations, score_range=(0.75, 1.0),
            seed=self.seed)
        pher_increase_50 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            limits=self.limits, weight=self.weight, iterations=iterations, score_range=(0.50, 1.0),
            seed=self.seed)
        pher_increase_25 = self._pherIncrease(
            n_ants=self.n_ants, pher_init=self.pher_init, evaporation=self.evaporation,
            limits=self.limits, weight=self.weight, iterations=iterations, score_range=(0.25, 1.0),
            seed=self.seed)
        pher_decrease = self._pherDecrease(
            pher_init=self.pher_init, evaporation=self.evaporation, limits=self.limits,
            iterations=iterations)

        fig, axes = plt.subplots(2, figsize=figsize)
        axes[0].plot(iterations, pher_increase_100, label='Scores [0.99-1.0]', color='#641E16',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_75, label='Scores [0.75-1.0]', color='#A93226',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_50, label='Scores [0.50-1.0]', color='#D98880',
                     linewidth=linewidth)
        axes[0].plot(iterations, pher_increase_25, label='Scores [0.25-1.0]', color='#F2D7D5',
                     linewidth=linewidth)
        axes[1].plot(iterations, pher_decrease, color='#2980B9', linewidth=linewidth)

        axes[0].set_title('Pheromone increase', fontsize=title_size)
        axes[0].spines['top'].set_visible(False)
        axes[0].spines['right'].set_visible(False)
        axes[0].legend()
        axes[1].set_title('Pheromone decrease', fontsize=title_size)
        axes[1].spines['top'].set_visible(False)
        axes[1].spines['right'].set_visible(False)

        if save_plot is not None:
            plt.savefig(save_plot, dpi=150)

        plt.show()

    @staticmethod
    def _pherIncrease(n_ants: int, pher_init: float, evaporation: float, limits: tuple, weight: float,
                      iterations: list, score_range: tuple, seed: int):
        """ Simulate the pheromone update (increasing the values) according to the MMAS update
        strategy. """
        if seed is not None:
            np.random.seed(seed)
        pher_increase = []
        pher = pher_init
        for it in iterations:
            scores = np.random.uniform(low=score_range[0], high=score_range[1], size=n_ants)
            for score in scores:
                pher = (1 - evaporation) * pher + score * weight
            if pher < limits[0]:
                pher = limits[0]
            if pher > limits[1]:
                pher = limits[1]
            pher_increase.append(pher)

        return pher_increase

    @staticmethod
    def _pherDecrease(pher_init: float, evaporation: float, limits: tuple, iterations: list):
        """ Simulate the pheromone update (decreasing the values) according to the MMAS update
        strategy. """
        pher_decrease = []
        pher = pher_init
        for it in iterations:
            pher = (1 - evaporation) * pher
            if pher < limits[0]:
                pher = limits[0]
            if pher > limits[1]:
                pher = limits[1]
            pher_decrease.append(pher)

        return pher_decrease



from setuptools import find_packages, setup
import numpy
from Cython.Build import cythonize


with open("README.md", 'r') as f:
    long_description = f.read()


# python setup.py build_ext -i
setup(name='antco',
      version='0.1.7',
      license='BSD3',
      description='Ant Colony Optimization framework',
      author='Fernando García Gutiérrez',
      author_email='fegarc05@ucm.es',
      url='https://github.com/FernandoGaGu/Ant-Colony-Optimisation/',                           
      download_url='https://github.com/FernandoGaGu/Ant-Colony-Optimisation/archive/refs/tags/0.1.tar.gz',
      install_requires=[
                'numpy', 
                'joblib', 
                'cython', 
                'matplotlib', 
                'deap', 
                'seaborn', 
                'networkx'
            ],
      keywords=['Optimisation', 'Ant Colony Optimisation', 'Metaheuristic', 'Algorithms'],
      packages=find_packages(),
      ext_modules=cythonize([
                './antco/c_metrics.pyx', 
                './antco/c_ntools.pyx',
                './antco/c_pheromone.pyx',
                './antco/c_policy.pyx',
                './antco/c_utils.pyx']),
      include_dirs=[numpy.get_include()],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',
          'License :: OSI Approved :: BSD License', 
          'Programming Language :: Python :: 3', 
          'Programming Language :: Python :: 3.4', 
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',],
      )



# Module that groups the evolutionary-based strategies that can be incorporated as a hybrid
# strategy in the algorithms.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import matplotlib.pyplot as plt
import random
import multiprocessing as mp
from deap import base
from deap import algorithms
from deap import creator
from deap import tools
from .base import MetaHeuristic
from ..optim import ObjectiveFunction
from ..ant import Ant


class PermutationGA(MetaHeuristic):
    """
    Class implementing a permutation-based evolutionary strategy. This class allows to optimise the
    permutations given by the ants, for example for the traveller salesman problem: each ant selects
    a route, the best routes are received by this algorithm which is in charge of refining them.
    The new routes refined by the evolutionary strategy are used to update the values of the
    pheromone matrix.

    The evolutionary algorithm executed will correspond to the one implemented in
    deap.algorithms.eaSimple. The user must specify the cost function used to perform the mapping
    from genotype (in the form of permutation) to phenotype and evaluate its quality.

    Parameters
    ----------
    antco_objective: antco.optim.ObjectiveFunction
        Objective function defined using the antco.optim.ObjectiveFunction interface.

    genetic_objective: callable
        Objective function that will receive each individual (encoded as a list of integers without
        repetition) and will return a tuple where the first element will be the scalar value
        associated with that genotype.

    best_ants: int
        Number of the best ants to be passed to the hybrid strategy.

    population_size: int
        Genetic algorithm population size.

    crossover_prob: float
        Genetic algorithm cross-over probability.

    mutation_prob: float
        Genetic algorithm mutation probability.

    individual_mutation_prob: float
        Genetic algorithm genotype position mutation probability.

    generations: int
        Genetic algorithm number of generations.

    tournsize: int
        Genetic algorithm number of individuals selected for tournament selection.

    hof: int, default=None
        Genetic algorithm elite (Hall of Fame) used in deap.algorithms.eaSimple

    n_jobs: int, default=1
        Number of processes running in parallel.

    genetic_objective_args: dict, default=None
        If the genetic objective function requires additional parameters these must be passed to
        the constructor in the form of a dictionary where the name of the parameter received in the
        objective function must correspond to the key and the value passed to the value.

    display_convergence: bool, default=False
        Parameter indicating whether to display the convergence graphs of the evolutionary
        algorithm at each iteration, useful for debugging and hyperparameter tuning purposes.
        Defaults to off.

    add_to_old: bool, default=False
        Parameter indicating whether to add the new solution to the set of ants or to only consider
        the number of ants returned by this strategy for the update of the pheromone matrix. In the
        latter case, the default behaviour, if the number of ants exploring the network is for
        example 100 and the parameter best_ants is 10, the number of ants that will be available
        for the pheromone matrix update will be 10. If the parameter is specified as True, the
        number of ants available will be 100 + 10.
    """
    def __init__(self, antco_objective: ObjectiveFunction, genetic_objective: callable,
                 best_ants: int, population_size: int, crossover_prob: float, mutation_prob: float,
                 individual_mutation_prob: float, generations: int, tournsize: int = 2,
                 hof: int = None, n_jobs: int = 1, genetic_objective_args: dict = None,
                 display_convergence: bool = False, add_to_old: bool = False):

        super(PermutationGA, self).__init__(antco_objective, add_to_old)

        self._best_ants = best_ants
        self._generations = generations
        self._mutation_prob = mutation_prob
        self._indpb = individual_mutation_prob
        self._crossover_prob = crossover_prob
        self._population_size = population_size
        self._display_convergence = display_convergence
        self._toolbox = base.Toolbox()

        # Create fitness function
        creator.create('FitnessMax', base.Fitness, weights=(1.0,))

        # Register GA elements
        if genetic_objective_args is None:
            self._toolbox.register('evaluate', genetic_objective)
        else:   # Pass fixed arguments to the cost function
            self._toolbox.register('evaluate', genetic_objective, **genetic_objective_args)
        self._toolbox.register('select', tools.selTournament, tournsize=tournsize)
        self._toolbox.register('mate', tools.cxOrdered)
        self._toolbox.register('mutate', tools.mutShuffleIndexes, indpb=individual_mutation_prob)

        # Create Hall of Fame
        self._hof = tools.HallOfFame(hof) if hof is not None else hof

        # Create stats
        if self._display_convergence:
            self._stats = tools.Statistics(lambda ind: ind.fitness.values)
            self._stats.register('max', np.max)
            self._stats.register('mean', np.mean)
        else:
            self._stats = None

        # Parallel execution
        if n_jobs > 1 or n_jobs == -1:
            pool = mp.Pool(mp.cpu_count() if n_jobs == -1 else n_jobs)
            self._toolbox.register("map", pool.map)

    def __repr__(self):
        return f'PermutationGA(best_ants={self._best_ants}, generations={self._generations}, ' \
               f'population_size={self._population_size}, mutation_prob={self._mutation_prob}, ' \
               f'crossover_prob={self._crossover_prob}, individual_mutation_prob={self._indpb})'

    def __str__(self):
        return self.__repr__()

    def optimise(self, ants: list, scores: list) -> tuple:
        # Get the index associated with the best ants
        best_ants = np.argsort(scores)[::-1][:self._best_ants]

        # Initialisation of the genotypes of the individuals using the best routes found by the ACO
        initial_genotypes = [ants[idx].visited_nodes.tolist() for idx in best_ants]
        self.create_permutation_ind(self._toolbox, creator.FitnessMax, initial_genotypes)

        # Create initial population
        self._toolbox.register('population', tools.initRepeat, list, self._toolbox.individual)
        population = self._toolbox.population(n=self._population_size)

        # Execute algorithm
        if self._display_convergence:
            population, logbook = algorithms.eaSimple(
                population, self._toolbox, cxpb=self._crossover_prob, mutpb=self._mutation_prob,
                ngen=self._generations, stats=self._stats, halloffame=self._hof, verbose=False)

            plt.plot(logbook.select('max'), color='red')
            plt.plot(logbook.select('mean'), color='green')
            plt.xlabel('Generation')
            plt.ylabel('Max / Average Fitness')
            plt.title('Max and Average fitness over Generations')
            plt.show()
        else:
            population, logbook = algorithms.eaSimple(
                population, self._toolbox, cxpb=self._crossover_prob, mutpb=self._mutation_prob,
                ngen=self._generations, stats=self._stats, halloffame=self._hof, verbose=False)
        # Unregister elements
        del creator.Individual
        self._toolbox.unregister('individual')
        self._toolbox.unregister('attr_permutation')

        return processPopulation(best_ants=self._best_ants,
                                 ant_prototype=ants[0],
                                 population=population,
                                 hof=self._hof)

    @classmethod
    def create_permutation_ind(cls, toolbox: base.Toolbox, fitness_function: callable,
                               initial_values: list = None, individual_size: int = None):
        """
        Method that allows to create and register (following the guidelines defined in DEAP) the
        genotype of the individuals (registered as 'Individual') and the generating function of
        individuals (registered as 'individual').

        Parameters
        ----------
        toolbox: base.Toolbox
            DEAP Toolbox instance.

        fitness_function: callable
            DEAP fitness function.

        initial_values: list, default=None
            List of list of initial genotypes used for the creation of the initial population, this
            allows incorporating a priori knowledge about better solutions and usually gives better
            results than random initialisation of the genotypes.

            If this parameter is not provided, it will be necessary to provide tha argument
            individual_size.

        individual_size: int, default=None
            Size of the individual genotype.

            If this parameter is not provided, it will be necessary to provide tha argument
            individual_size.

        Notes
        -----
        Parameters initial_values and individual_size cannot be provided at the same time.
        """
        assert (initial_values is None or individual_size is None) and \
               not (initial_values is None and individual_size is None), \
               'Either the initial_values or individual_size must be provided.'

        # Create from initial values
        if initial_values is not None:
            ind_generator = lambda initial_values: initial_values[
                random.randint(0, len(initial_values) - 1)]
            toolbox.register('attr_permutation', ind_generator, initial_values)
        # Create randomly
        else:
            toolbox.register('attr_permutation', random.sample, range(individual_size), individual_size)

        creator.create('Individual', list, fitness=fitness_function)
        toolbox.register('individual', tools.initIterate, creator.Individual, toolbox.attr_permutation)


class SetGA(MetaHeuristic):
    """
    Class implementing a integer-based evolutionary strategy. This class allows optimisation to be
    performed on a genotype of non-repeating integer values of variable length.

    Parameters
    ----------
    antco_objective: antco.optim.ObjectiveFunction
        Objective function defined using the antco.optim.ObjectiveFunction interface.

    genetic_objective: callable
        Objective function that will receive each individual (encoded as a list of integers without
        repetition) and will return a tuple where the first element will be the scalar value
        associated with that genotype.

    best_ants: int
        Number of the best ants to be passed to the hybrid strategy.

    population_size: int
        Genetic algorithm population size.

    crossover_prob: float
        Genetic algorithm cross-over probability.

    mutation_prob: float
        Genetic algorithm mutation probability.

    generations: int
        Genetic algorithm number of generations.

    tournsize: int
        Genetic algorithm number of individuals selected for tournament selection.

    hof: int, default=None
        Genetic algorithm elite (Hall of Fame) used in deap.algorithms.eaSimple

    n_jobs: int, default=1
        Number of processes running in parallel.

    genetic_objective_args: dict, default=None
        If the genetic objective function requires additional parameters these must be passed to
        the constructor in the form of a dictionary where the name of the parameter received in the
        objective function must correspond to the key and the value passed to the value.

    display_convergence: bool, default=False
        Parameter indicating whether to display the convergence graphs of the evolutionary
        algorithm at each iteration, useful for debugging and hyperparameter tuning purposes.
        Defaults to off.

    add_to_old: bool, default=False
        Parameter indicating whether to add the new solution to the set of ants or to only consider
        the number of ants returned by this strategy for the update of the pheromone matrix. In the
        latter case, the default behaviour, if the number of ants exploring the network is for
        example 100 and the parameter best_ants is 10, the number of ants that will be available
        for the pheromone matrix update will be 10. If the parameter is specified as True, the
        number of ants available will be 100 + 10.
    """
    def __init__(self, antco_objective: ObjectiveFunction, genetic_objective: callable,
                 best_ants: int, population_size: int, crossover_prob: float, mutation_prob: float,
                 generations: int, tournsize: int = 2, hof: int = None, n_jobs: int = 1,
                 genetic_objective_args: dict = None, display_convergence: bool = False,
                 add_to_old: bool = False):

        super(SetGA, self).__init__(antco_objective, add_to_old)

        self._best_ants = best_ants
        self._generations = generations
        self._mutation_prob = mutation_prob
        self._crossover_prob = crossover_prob
        self._population_size = population_size
        self._display_convergence = display_convergence
        self._toolbox = base.Toolbox()

        # Create fitness function
        creator.create('FitnessMax', base.Fitness, weights=(1.0,))

        # Register GA elements
        if genetic_objective_args is None:
            self._toolbox.register('evaluate', genetic_objective)
        else:   # Pass fixed arguments to the cost function
            self._toolbox.register('evaluate', genetic_objective, **genetic_objective_args)
        self._toolbox.register('select', tools.selTournament, tournsize=tournsize)
        self._toolbox.register('mate', cxSet)

        # Create Hall of Fame
        self._hof = tools.HallOfFame(hof) if hof is not None else hof

        # Create stats
        if self._display_convergence:
            self._stats = tools.Statistics(lambda ind: ind.fitness.values)
            self._stats.register('max', np.max)
            self._stats.register('mean', np.mean)
        else:
            self._stats = None

        # Parallel execution
        if n_jobs > 1 or n_jobs == -1:
            pool = mp.Pool(mp.cpu_count() if n_jobs == -1 else n_jobs)
            self._toolbox.register("map", pool.map)

    def __repr__(self):
        return f'SetGA(best_ants={self._best_ants}, generations={self._generations}, ' \
               f'population_size={self._population_size}, mutation_prob={self._mutation_prob}, ' \
               f'crossover_prob={self._crossover_prob})'

    def __str__(self):
        return self.__repr__()

    def optimise(self, ants: list, scores: list) -> tuple:
        # Get the index associated with the best ants
        best_ants = np.argsort(scores)[::-1][:self._best_ants]
        possible_values = np.unique([   # Select unique nodes
            node for idx in best_ants for node in ants[idx].visited_nodes]).tolist()

        # Register mutation operator
        self._toolbox.register('mutate', mutSet, possible_values=possible_values)

        # Initialisation of the genotypes of the individuals using the list of possible values
        self.create_set_ind(self._toolbox, creator.FitnessMax, possible_values)

        # Create initial population
        self._toolbox.register('population', tools.initRepeat, list, self._toolbox.individual)
        population = self._toolbox.population(n=self._population_size)

        # Execute algorithm
        if self._display_convergence:
            population, logbook = algorithms.eaSimple(
                population, self._toolbox, cxpb=self._crossover_prob, mutpb=self._mutation_prob,
                ngen=self._generations, stats=self._stats, halloffame=self._hof, verbose=False)

            plt.plot(logbook.select('max'), color='red')
            plt.plot(logbook.select('mean'), color='green')
            plt.xlabel('Generation')
            plt.ylabel('Max / Average Fitness')
            plt.title('Max and Average fitness over Generations')
            plt.show()
        else:
            population, logbook = algorithms.eaSimple(
                population, self._toolbox, cxpb=self._crossover_prob, mutpb=self._mutation_prob,
                ngen=self._generations, stats=self._stats, halloffame=self._hof, verbose=False)

        # Unregister elements
        del creator.Individual
        self._toolbox.unregister('individual')
        self._toolbox.unregister('attr_set')
        self._toolbox.unregister('mutate')

        return processPopulation(best_ants=self._best_ants,
                                 ant_prototype=ants[0],
                                 population=population,
                                 hof=self._hof)

    @classmethod
    def create_set_ind(cls, toolbox: base.Toolbox, fitness_function: callable,
                       possible_values: list):
        """
        Method to create the genotype of the individuals represented by a Set from the list of
        possible values received as an argument.

        Parameters
        ----------
        toolbox: base.Toolbox
            DEAP Toolbox instance.

        fitness_function: callable
            DEAP fitness function.

        possible_values: list
            List of possible values to insert into the individual.
        """
        creator.create('Individual', set, fitness=fitness_function)
        toolbox.register('attr_set', initSetGenotype, possible_values=list(possible_values))
        toolbox.register('individual', tools.initIterate, creator.Individual, toolbox.attr_set)


def cxSet(ind1, ind2):
    """
    Apply a crossover operation on input sets. The first child is the intersection of the two sets,
    the second child is the difference of the two sets.

    Parameters
    ----------
    ind1: Set
        Parent 1.

    ind2: Set
        Parent 1

    Returns
    -------
    :tuple
        :[0]: Set
            Child 1.
        :[1]: Set
            Child 2.
    """
    temp = set(ind1)                # Used in order to keep type
    ind1 &= ind2                    # Intersection (inplace)
    ind2 ^= temp                    # Symmetric Difference (inplace)
    return ind1, ind2


def mutSet(individual, possible_values: list):
    """
    Mutation that pops or add an element.

    Parameters
    ----------
    individual: Set
        Individual to be mutated.

    possible_values: list
        List of possible values to insert into the individual.

    Returns
    -------
    :tuple
        :[0]: Set
            Mutated individual.
    """
    if random.random() < 0.5:
        if len(individual) > 0:     # We cannot pop from an empty set
            individual.remove(random.choice(sorted(tuple(individual))))
    else:
        individual.add(random.choice(list(possible_values)))
    return individual,


def initSetGenotype(possible_values: list):
    """
    Function to initialise the genotype of individuals represented as a Set from a list of possible
    values.

    Parameters
    ----------
    possible_values: list
        List of possible values to insert into the individual.

    Returns
    -------
    :list
        Individual genotype.

    Notes
    -----
    There must be at least two different values in possible_values.
    """
    num_elements = random.randint(2, len(possible_values) - 1)
    return np.random.choice(possible_values, size=num_elements, replace=False).tolist()


def processPopulation(best_ants: int, ant_prototype: Ant, population: list,
                      hof: tools.HallOfFame or None = None) -> tuple:
    """ Function in responsible for creating the ants to be returned by the metaheuristic strategies
    according to the interface defined in antco.hybrid.base.Metaheuristic from the populations of
    individuals of the evolutionary strategies. """

    if hof is None:  # Create Ants (when Hall of Fame hasn't been specified)
        improved_ants = [ant_prototype.new() for _ in range(best_ants)]
        improved_scores = [-float('inf') for _ in range(best_ants)]
        fitness_values = [ind.fitness.values[0] for ind in population]
        best_fitness = np.argsort(fitness_values)[::-1][:best_ants]
        for idx, ind_idx in enumerate(best_fitness):
            improved_ants[idx].visited_nodes = list(population[ind_idx])
            improved_scores[idx] = fitness_values[idx]
    else:  # Create Ants (using Hall of Fame )
        improved_ants = [ant_prototype.new() for _ in range(len(hof.items))]
        improved_scores = [-float('inf') for _ in range(len(hof.items))]
        for idx, solution in enumerate(hof.items):
            improved_ants[idx].visited_nodes = list(solution)
            improved_scores[idx] = solution.fitness.values[0]

    return improved_ants, improved_scores



# Module that groups the greedy search strategies that can be incorporated as a hybrid strategy in
# the algorithms.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import joblib
import networkx as nx
from .base import MetaHeuristic
from ..optim import ObjectiveFunction
from ..c_ntools import getValidPaths


class GreedySearchOP(MetaHeuristic):
    """
    Class defining a greedy search strategy for the selection of optimal paths.This class allows to
    optimise the path given by the ants following the following algorithm:

    ----------------------------------------------------------------------------------------------
        Input: G (Graph defining the network structure)
        Input: max_depth (Maximum depth of the exploration)
        Input: ant (Ant given by the ACO algorithm)

        1. solution <- ant.visited_nodes
        2. solution_score <- evaluate(solution)
        3. depth = 0
        4. While depth < max_depth:
        5.     improvement = False
               # Get those nodes that when removed do not generate two disconnected sub-networks
        6.     nodes <- getReplaceable(solution)
        7.     for node in nodes:
        8.         new_solution <- remove(solution, node)  # Eliminate node from solution
                   # Nodes that can be added to the network according to the network structure
        9.         available_nodes <- getValidNodes(new_solution, G)
        10.         for new_node in available_nodes:
        11.             new_solution_temp <- add(new_solution, new_node)
        12.            new_score <- evaluate(new_solution_temp)
        13.            if new_score > solution_score:
        14.                solution <- new_solution_temp
        15.                solution_score <- new_score
                           improvement = True
        16.    if not improvement: break
        17.    depth++
        18. Process and return solution
    ----------------------------------------------------------------------------------------------

    Important note: This class is not of general purpose, for certain types of problems the
    strategy may not make sense. Also, depending on the network to be explored, it may be
    inefficient.

    Important note (2): See parameter add_to_old.

    Parameters
    ----------
    antco_objective: antco.optim.ObjectiveFunction
        Objective function defined using the antco.optim.ObjectiveFunction interface.

    best_ants: int
        Number of the best ants to be passed to the hybrid strategy.

    objective: callable
        Objective function that will receive each solution (encoded as a list of integers without
        repetition) and will return a scalar value evaluating the solution.

    adj_matrix: np.ndarray
        Adjacency matrix defining the network to be explored.

    n_jobs: int, default=1
        Number of processes running in parallel.

    objective_args: dict, default=None
        If the objective function requires additional parameters these must be passed to the
        constructor in the form of a dictionary where the name of the parameter received in the
        objective function must correspond to the key and the value passed to the value.

    add_to_old: bool, default=False
        Parameter indicating whether to add the new solution to the set of ants or to only consider
        the number of ants returned by this strategy for the update of the pheromone matrix. In the
        latter case, the default behaviour, if the number of ants exploring the network is for
        example 100 and the parameter best_ants is 10, the number of ants that will be available
        for the pheromone matrix update will be 10. If the parameter is specified as True, the
        number of ants available will be 100 + 10.
    """
    def __init__(self, antco_objective: ObjectiveFunction, best_ants: int, depth: int,
                 objective: callable, adj_matrix: np.ndarray, n_jobs: int = 1,
                 objective_args: dict = None, add_to_old: bool = False):
        super(GreedySearchOP, self).__init__(antco_objective, add_to_old)

        self._best_ants = best_ants
        self._depth = depth
        self._adj_matrix = adj_matrix
        self._objective = objective
        self._objective_args = {} if objective_args is None else objective_args
        self._n_jobs = n_jobs

    def __repr__(self):
        return f'GreedySearchOP(best_ants={self._best_ants}, depth={self._depth}, ' \
               f'n_jobs={self._n_jobs})'

    def __str__(self):
        return self.__repr__()

    def optimise(self, ants: list, scores: list) -> tuple:
        # Get the index associated with the best ants
        best_ants_indices = np.argsort(scores)[::-1][:self._best_ants]
        improved_solutions = []
        improved_scores = []

        if self._n_jobs > 1:
            solution_score = joblib.Parallel(n_jobs=self._n_jobs, backend='loky')(
                joblib.delayed(greedySearch)(
                    self.evaluate.getVisitedNodes(ants[idx]), self._depth, self._adj_matrix,
                    self._objective, self._objective_args, scores[idx]) for idx in best_ants_indices)

            for (new_solution, new_score) in solution_score:
                improved_solutions.append(new_solution)
                improved_scores.append(new_score)

        else:  # Serial execution
            for idx in best_ants_indices:
                new_solution, new_score = greedySearch(
                    nodes=self.evaluate.getVisitedNodes(ants[idx]),
                    depth=self._depth,
                    adj_matrix=self._adj_matrix,
                    objective=self._objective,
                    objective_args=self._objective_args,
                    best_score=scores[idx])

                improved_solutions.append(new_solution)
                improved_scores.append(new_score)

        # Create ants
        improved_ants = [ants[0].new() for _ in range(self._best_ants)]
        for idx, solution in enumerate(improved_solutions):
            improved_ants[idx].visited_nodes = solution

        return improved_ants, improved_scores


def greedySearch(nodes: list or np.ndarray, depth: int, adj_matrix: np.ndarray,
                 objective: callable, objective_args: dict, best_score: float):
    """
    Greedy search strategy used in GreedySearchOP.

    Parameters
    ----------
    nodes: list or np.ndarray (solution_nodes)
        Nodes which make up the solution to be improved.

    depth: int
        Maximum depth of the solution tree explored.

    adj_matrix: np.ndarray (nodes, nodes)
        Adjacency matrix defining the structure of the network to be explored.

    objective: callable
        Objective function to be maximised used for the evaluation of solutions.

    objective_args: dict
        Additional parameters that will be passed to the objective function.

    best_score: float
        Initial score associated to the initial solution.

    Returns
    -------
    :tuple
        Tuple where the first element correspond to the improved solution or the solution received
        if it could not be improved, and the second argument to the score associated with the
        solution.
    """
    graph = nx.from_numpy_array(adj_matrix)

    best_solution = None
    current_depth = 0
    while current_depth < depth:
        if current_depth == 0:
            nodes_to_explore = [n for n in nodes]
        else:
            if best_solution is None: break
            nodes_to_explore = best_solution

        # Get those nodes that when removed do not generate two disconnected sub-networks
        repl_nodes = getReplaceable(nodes_to_explore, graph)

        # Replace each possible node
        for n1 in repl_nodes:
            # Solution in which one of the nodes has been removed
            pruned_solution = [node for node in nodes_to_explore]
            pruned_solution.pop(pruned_solution.index(n1))

            # Get a list of all nodes that can be added to the sub-network according to the network
            # structure
            substitutes = []
            for n2 in pruned_solution:
                substitutes.append(
                    getValidPaths(n2, np.array(pruned_solution), adj_matrix))

            # Getting the unique node
            substitutes = list(set([e for lt in substitutes for e in lt if not e == n1]))

            for n2 in substitutes:
                new_score = objective(pruned_solution + [n2], **objective_args)
                if new_score > best_score:
                    best_solution = pruned_solution + [n2]
                    best_score = new_score
        current_depth += 1

    if best_solution is None:
        return nodes, best_score

    return best_solution, best_score


def getReplaceable(nodes: np.ndarray, graph: nx.Graph) -> list:
    """
    Function that returns the nodes from the nodes argument, that can be removed from the network
    without generating two disconnected sub-networks.

    Parameters
    ----------
    nodes: np.ndarray
        Nodes composing the sub-graph whose nodes are to be examined.

    graph: nx.Graph
        Graph defining the structure of the network to be explored.

    Returns
    -------
    :list
        Nodes that can be replaced without generating two disconnected sub-networks.
    """
    replaceable_nodes = []
    for node in nodes:
        selected_nodes = [n for n in nodes if n != node]
        subgraph = graph.subgraph(selected_nodes)
        if nx.is_connected(subgraph):
            replaceable_nodes.append(node)

    return replaceable_nodes




# Module containing the experimental features of the framework that are still under development.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import sys
import random
import warnings
from copy import deepcopy

from ..report import Report
from ..ant import fixedPositions, deleteInitialPosition
from ..base import DecaySchedule
from ..aco import ACO, evaluateAnts, generatePathsACS
from ..algorithm import updateReportWithBest

warnings.filterwarnings('error')


def antColonySystem(aco_obj: ACO, scores_decay: DecaySchedule = None,
                    evaporation_decay: DecaySchedule = None, report: Report = None,
                    save_pheromones: bool = True, verbose: bool = True) -> Report:
    """
    Function that executes the Ant Colony System algorithm based on

        Dorigo, M., & Gambardella, L. M. (1997). Ant colony system: a cooperative learning
        approach to the traveling salesman problem. IEEE Transactions on evolutionary computation,
        1(1), 53-66.


    Parameters
    ----------
    aco_obj: antco.ACO
        Initialised antco.ACO instance, for more information on this type of object use:
        help(antco.ACO)

    scores_decay: antco.optim.DecaySchedule subclass, default=None
        Instance of a antco.optim.DecaySchedule representing a decay schedule that will applied to
        the score values with which the pheromone values are updated in each iteration.
        For more info use help(antco.optim.DecaySchedule).

    evaporation_decay: antco.optim.DecaySchedule subclass, default=None
        Same as scores_decay, in this case used to update the evaporation rate value.
        For more info use help(antco.optim.DecaySchedule).

    report: antco.report.Report, default=None
        antco.report.Report instance, by default a report will be initialised.

    save_pheromones: bool, default=True
        Value indicating whether to store in the report the values of the pheromone matrix at each
        iteration. If True these may be retrieved by:

            >>> report.get('pheromones')

    verbose: bool, default=True
        Indicates whether to show convergence information during the execution of the algorithm.

    Returns
    -------
    :antco.report.Report
        Returns a antco.report.Report instance with information about the convergence of the
        algorithm. The convergence parameters can be accessed via:

        >>> report = antco.algorithm.basic(...)
        >>> report.get('min_cost')
        >>> report.get('mean_cost')
        >>> report.get('max_cost')

        For more info about Report instances use help(antco.report.Report).
    """
    # Get parameters from aco_obj
    seed = aco_obj.seed
    graph = aco_obj.graph
    iterations = aco_obj.iterations
    n_jobs = aco_obj.n_jobs
    ants = aco_obj.ants
    H = aco_obj.heuristic
    P = aco_obj.pheromones
    alpha = aco_obj.alpha
    beta = aco_obj.beta
    if aco_obj.precompute_heuristic:
        H = np.power(H, beta)
    pheromone_update_kw = aco_obj.pheromone_update_kw
    rho = pheromone_update_kw['rho']; del pheromone_update_kw['rho']  # Get evaporation parameter
    decay = pheromone_update_kw['decay']; del pheromone_update_kw['decay']   # Get decay parameter
    pher_init_val = aco_obj.pher_init_val
    updatePheromones = aco_obj.updatePheromones
    objectiveFunction = aco_obj.objectiveFunction
    tol = aco_obj.tol
    fixed = aco_obj.fixed_positions
    accessory_node = aco_obj.accessory_node
    Q = aco_obj.Q
    scaleScores = aco_obj.scaleScores

    if report is None:  # Initialize a Report instance (if not provided)
        report = Report({'BranchingFactor': {'lambda_values': [0.5]}})

    if seed is not None:  # Set random state
        random.seed(seed)
        np.random.seed(seed)

    if fixed and accessory_node is None:  # Put ants in fixed positions (if specified)
        fixedPositions(ants, graph)

    # If an accessory node has been specified, place the ants on that node
    if accessory_node is not None:
        for ant in ants:
            ant.setInitialPosition(accessory_node)

    # Pre-compute the array to store the ant scores
    ant_scores, norm_scores = np.empty(shape=len(ants)), np.empty(shape=len(ants))

    current_iteration = 1
    it_without_improvements = 0
    best_score = -float('inf')
    while current_iteration <= iterations:

        ants = generatePathsACS(   # Graph exploration
            ants=ants, graph=graph, H=H, P=P, alpha=alpha, beta=beta, decay=decay,
            pher_init=pher_init_val, Q=Q, exp_heuristic=False)

        # Evaluate ant paths using the objectiveFunction function (it will be maximized)
        ant_scores = evaluateAnts(ants=ants, objectiveFunction=objectiveFunction)

        # Update best score and save best solution
        new_best = updateReportWithBest(
            ants=ants, scores=ant_scores, best_score=best_score, report=report,
            iteration=current_iteration)

        if new_best > best_score:
            best_score = new_best
            it_without_improvements = 0
        else:
            it_without_improvements += 1

        # Scale scores
        if scaleScores is not None:
            norm_scores[:] = ant_scores[:]  # Copy scores
            norm_scores = scaleScores(norm_scores, best_score)
        else:
            norm_scores = ant_scores

        # Update pheromones according to the scores
        updatePheromones(
            paths=np.array([ant.adj_matrix for ant in ants], dtype=np.int8), P=P,
            ant_scores=norm_scores if scores_decay is None else norm_scores * scores_decay(current_iteration),
            rho=rho if evaporation_decay is None else rho * evaporation_decay(current_iteration),
            **pheromone_update_kw)

        if not fixed and accessory_node is None:  # Restart ants initial position
            deleteInitialPosition(ants)

        # Compute convergence statistics
        mean_scores = np.mean(ant_scores)
        min_score = np.min(ant_scores)
        max_score = np.max(ant_scores)
        report.save(current_iteration, mean_cost=mean_scores, max_cost=max_score)

        if save_pheromones:  # Save pheromone values
            report.save(current_iteration, pheromones=deepcopy(P))

        # Compute monitoring computeMetrics
        report.computeMetrics(current_iteration, P, graph)

        # After several generations without improvements, do an early stopping of the algorithm.
        if it_without_improvements > tol: break

        if verbose:
            sys.stdout.write('\rCost: Mean: %.4f (Min: %.4f Max: %.4f) (iteration %d)' %
                             (float(mean_scores), min_score, max_score, current_iteration))

        current_iteration += 1

    return report



# Module defining the necessary interfaces related to the optimisation process.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
from abc import ABCMeta, abstractmethod
from .ant import Ant


class ObjectiveFunction(object):
    """
    Object defining the cost function to be maximised by the ant optimisation algorithm.

    Methods
    -------
    getVisitedNodes(ant: antco.Ant) -> list
        Method which, given an ant, returns a list of nodes visited by that ant.

    Attributes
    ----------
    accessory_node: bool
        Returns True if an accessor node has been added in the pre-processing of an antco.ACO
        instance, otherwise return False.

    Abstract methods
    ----------------
    evaluate(ant: antco.Ant)
        Abstract method specific to the optimisation problem to be implemented by the user. This
        method must receive an antco.Ant instance and return a scalar value. Internal modifications
        of the instance received by the function will have no effect.
    """
    __metaclass__ = ABCMeta

    _accessory_node = False

    def __repr__(self):
        return f'ObjectiveFunction'

    def __str__(self):
        return self.__repr__()

    def __call__(self, ant_idx: tuple) -> tuple:
        return self.evaluate(ant_idx[0]), ant_idx[1]

    def getVisitedNodes(self, ant: Ant):
        if self._accessory_node:
            return ant.visited_nodes[1:]
        else:
            return ant.visited_nodes

    @property
    def accessory_node(self):
        return self._accessory_node

    @accessory_node.setter
    def accessory_node(self, present: bool):
        self._accessory_node = present

    @abstractmethod
    def evaluate(self, ant: Ant) -> float:
        """
        Method that evaluates the path travelled by a given ant in such a way that it returns a
        numerical value which will be maximised.

        Parameters
        ----------
        ant: Ant
            Ant encoding the path traversed by the graph.

        Returns
        -------
        :float
            Score associated with the ant.

        Notes
        ------
        Since the framework is optimised for parallel execution, modifications to the internal
        state of the Ant during the execution of ObjectiveFunction instances will have no effect
        on the internal state.
        """
        raise NotImplementedError





import numpy as np
from antco import Ant, randomInit, fixedPositions


def test():
    seed = 1997
    np.random.seed(seed)

    l_min = 2
    l_max = 4

    adjacency_matrix = np.array([
        [0, 1, 1, 0, 1, 1],
        [1, 0, 1, 1, 0, 0],
        [1, 1, 0, 0, 0, 1],
        [0, 1, 0, 0, 1, 1],
        [1, 0, 0, 1, 0, 0],
        [0, 0, 1, 1, 0, 0]], dtype=np.int8)

    # Check constructors
    _ = Ant(l_min, l_max, 'd')
    _ = Ant(l_min, l_max, 'directed')
    _ = Ant(l_min, l_max, 'u')
    _ = Ant(l_min, l_max, 'undirected')

    # Check limits
    l_ant = Ant(l_min, l_max, 'd')
    assert l_ant.max_length == l_max and l_ant.min_length == l_min, \
        'ERROR: antco.ant.max_length or antco.ant.min_length'

    # Check initialization
    # Directed
    d_ant = Ant(l_min, l_max, 'd')
    d_ant.initAdjMatrix(n_nodes=adjacency_matrix.shape[0])
    d_ant.setInitialPosition(randomInit(adjacency_matrix))
    assert d_ant.initial_position == 5, 'ERROR: antco.ant.randomInit() (Directed)'
    assert len(d_ant.visited_nodes) == 1, 'ERROR: antco.ant.getVisitedNodes'
    # Undirected
    u_ant = Ant(l_min, l_max, 'u')
    u_ant.initAdjMatrix(n_nodes=adjacency_matrix.shape[0])
    u_ant.setInitialPosition(randomInit(adjacency_matrix))
    u_ant._adjacency_matrix[1, 3] = 1
    assert len(u_ant.visited_nodes) == 2, 'ERROR: antco.ant.getVisitedNodes (Undirected)'

    # Check assignment workflow
    # Directed
    d_ant = Ant(l_min, l_max, 'd')
    d_ant.setInitialPosition(0)
    d_ant.adj_matrix = np.array([
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 1, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0]], dtype=np.int8)

    assert np.all(d_ant.visited_nodes == [0, 3, 2, 4, 1, 5]), \
        'ERROR: antco.ant.adj_matrix or antco.ant.getVisitedNodes (Directed)'
    assert not d_ant.is_valid, 'ERROR: antco.ant.is_valid'

    d_ant.adj_matrix = np.array([
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0]], dtype=np.int8)
    assert np.all(d_ant.visited_nodes == [0, 4, 1]), \
        'ERROR: antco.ant.adj_matrix or antco.ant.getVisitedNodes (Directed)'

    # Undirected
    u_ant = Ant(l_min, l_max, 'u')
    u_ant.setInitialPosition(0)
    u_ant.adj_matrix = np.array([
        [0, 1, 1, 0, 0, 0],
        [1, 0, 1, 1, 0, 0],
        [1, 1, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0]], dtype=np.int8)

    assert np.all(u_ant.visited_nodes == [0, 1, 2]), \
        'ERROR: antco.ant.adj_matrix or antco.ant.getVisitedNodes (Undirected)'

    u_ant.adj_matrix = np.array([
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0]], dtype=np.int8)
    assert len(u_ant.visited_nodes) == 1, \
        'ERROR: antco.ant.adj_matrix or antco.ant.getVisitedNodes (Undirected)'

    # Random initial position
    struct = np.random.randint(2, size=u_ant.adj_matrix.shape, dtype=np.int8)
    u_ant.setInitialPosition(randomInit(struct))
    assert u_ant.initial_position == 0, 'ERROR: antco.ant.randomInit'

    # Ant fixed positions
    ants = [Ant(l_min=l_min, l_max=l_max, graph_type='u') for _ in range(10)]
    fixedPositions(ants, struct)
    covered_positions = [ant.initial_position for ant in ants[:6]]
    assert np.sort(covered_positions).tolist() == [0, 1, 2, 3, 4, 5], \
        'ERROR: antco.ant.fixedPositions'

    print('SUCCESSFUL TEST: antco.ant.Ant')
